from robyn import Robyn

app = Robyn(__file__)


@app.get("/")
def index():
    return "Hello World!"


if __name__ == "__main__":
    # create a configured "Session" class
    app.start(host="0.0.0.0", port=8080)



from dataclasses import dataclass
from typing import Dict, NewType, Optional, TypedDict


@dataclass
class Directory:
    route: str
    directory_path: str
    show_files_listing: bool
    index_file: Optional[str]

    def as_list(self):
        return [
            self.route,
            self.directory_path,
            self.show_files_listing,
            self.index_file,
        ]


PathParams = NewType("PathParams", Dict[str, str])
Method = NewType("Method", str)
FormData = NewType("FormData", Dict[str, str])
Files = NewType("Files", Dict[str, bytes])
IPAddress = NewType("IPAddress", Optional[str])


class JSONResponse(TypedDict):
    """
    A type alias for openapi response bodies. This class should be inherited by the response class type definition.
    """

    pass


class Body:
    """
    A type alias for openapi request bodies. This class should be inherited by the request body class annotation.
    """

    pass


__all__ = ["JSONResponse", "Body"]



import platform
import socket


def get_network_host():
    # windows doesn't support 0.0.0.0
    if platform.system() == "Windows":
        hostname = socket.gethostname()
        ip_address = socket.gethostbyname(hostname)
        return ip_address
    else:
        return "0.0.0.0"



import os
import pathlib

import pytest

from robyn.env_populator import load_vars, parser


# this tests if a connection can be made to the server with the correct port imported from the env file
@pytest.mark.benchmark
def test_env_population(test_session, env_file):
    path = pathlib.Path(__file__).parent
    env_path = path / "robyn.env"
    load_vars(variables=parser(config_path=env_path))
    PORT = os.environ["ROBYN_PORT"]
    HOST = os.environ["ROBYN_HOST"]
    assert PORT == "8080"
    assert HOST == "127.0.0.1"



import orjson


def jsonify(input_dict: dict) -> str:
    """
    This function serializes input dict to a json string

    Attributes:
        input_dict dict: response of the function
    """
    output_binary = orjson.dumps(input_dict)
    output_str = output_binary.decode("utf-8")
    return output_str



import inspect
import json
import typing
from dataclasses import asdict, dataclass, field
from importlib import resources
from inspect import Signature
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, TypedDict

from robyn.responses import html
from robyn.robyn import QueryParams, Response
from robyn.types import Body


class str_typed_dict(TypedDict):
    key: str
    value: str


@dataclass
class Contact:
    """
    The contact information for the exposed API.
    (https://swagger.io/specification/#contact-object)

    @param name: Optional[str] The identifying name of the contact person/organization.
    @param url: Optional[str] The URL pointing to the contact information. This MUST be in the form of a URL.
    @param email: Optional[str] The email address of the contact person/organization. This MUST be in the form of an email address.
    """

    name: Optional[str] = None
    url: Optional[str] = None
    email: Optional[str] = None


@dataclass
class License:
    """
    The license information for the exposed API.
    (https://swagger.io/specification/#license-object)

    @param name: Optional[str] The license name used for the API.
    @param url: Optional[str] A URL to the license used for the API. This MUST be in the form of a URL.
    """

    name: Optional[str] = None
    url: Optional[str] = None


@dataclass
class Server:
    """
    An array of Server Objects, which provide connectivity information to a target server. If the servers property is
    not provided, or is an empty array, the default value would be a Server Object with a url value of /.
    (https://swagger.io/specification/#server-object)

    @param url: str A URL to the target host. This URL supports Server Variables and MAY be relative,
    to indicate that the host location is relative to the location where the OpenAPI document is being served.
    Variable substitutions will be made when a variable is named in {brackets}.
    @param description: Optional[str] An optional string describing the host designated by the URL.
    """

    url: str
    description: Optional[str] = None


@dataclass
class ExternalDocumentation:
    """
    Additional external documentation for this operation.
    (https://swagger.io/specification/#external-documentation-object)

    @param description: Optional[str] A description of the target documentation.
    @param url: Optional[str] The URL for the target documentation.
    """

    description: Optional[str] = None
    url: Optional[str] = None


@dataclass
class Components:
    """
    Additional external documentation for this operation.
    (https://swagger.io/specification/#components-object)

    @param schemas: Optional[Dict[str, Dict]] An object to hold reusable Schema Objects.
    @param responses: Optional[Dict[str, Dict]] An object to hold reusable Response Objects.
    @param parameters: Optional[Dict[str, Dict]] An object to hold reusable Parameter Objects.
    @param examples: Optional[Dict[str, Dict]] An object to hold reusable Example Objects.
    @param requestBodies: Optional[Dict[str, Dict]] An object to hold reusable Request Body Objects.
    @param securitySchemes: Optional[Dict[str, Dict]] An object to hold reusable Security Scheme Objects.
    @param links: Optional[Dict[str, Dict]] An object to hold reusable Link Objects.
    @param callbacks: Optional[Dict[str, Dict]] An object to hold reusable Callback Objects.
    @param pathItems: Optional[Dict[str, Dict]] An object to hold reusable Callback Objects.
    """

    schemas: Optional[Dict[str, Dict]] = field(default_factory=dict)
    responses: Optional[Dict[str, Dict]] = field(default_factory=dict)
    parameters: Optional[Dict[str, Dict]] = field(default_factory=dict)
    examples: Optional[Dict[str, Dict]] = field(default_factory=dict)
    requestBodies: Optional[Dict[str, Dict]] = field(default_factory=dict)
    securitySchemes: Optional[Dict[str, Dict]] = field(default_factory=dict)
    links: Optional[Dict[str, Dict]] = field(default_factory=dict)
    callbacks: Optional[Dict[str, Dict]] = field(default_factory=dict)
    pathItems: Optional[Dict[str, Dict]] = field(default_factory=dict)


@dataclass
class OpenAPIInfo:
    """
    Provides metadata about the API. The metadata MAY be used by tooling as required.
    (https://swagger.io/specification/#info-object)

    @param title: str The title of the API.
    @param version: str The version of the API.
    @param description: Optional[str] A description of the API.
    @param termsOfService: Optional[str] A URL to the Terms of Service for the API.
    @param contact: Contact The contact information for the exposed API.
    @param license: License The license information for the exposed API.
    @param servers: list[Server] An list of Server objects representing the servers.
    @param externalDocs: Optional[ExternalDocumentation] Additional external documentation.
    @param components: Components An element to hold various schemas for the document.
    """

    title: str = "Robyn API"
    version: str = "1.0.0"
    description: Optional[str] = None
    termsOfService: Optional[str] = None
    contact: Contact = field(default_factory=Contact)
    license: License = field(default_factory=License)
    servers: List[Server] = field(default_factory=list)
    externalDocs: Optional[ExternalDocumentation] = field(default_factory=ExternalDocumentation)
    components: Components = field(default_factory=Components)


@dataclass
class OpenAPI:
    """
    This is the root object of the OpenAPI document.

    @param info: OpenAPIInfo Provides metadata about the API.
    @param openapi_spec: dict The content of openapi.json as a dict
    """

    info: OpenAPIInfo = field(default_factory=OpenAPIInfo)
    openapi_spec: dict = field(init=False)
    openapi_file_override: bool = False  # denotes whether there is an override or not.

    def __post_init__(self):
        """
        Initializes the openapi_spec dict
        """
        if self.openapi_file_override:
            return

        self.openapi_spec = {
            "openapi": "3.1.0",
            "info": asdict(self.info),
            "paths": {},
            "components": asdict(self.info.components),
            "servers": [asdict(server) for server in self.info.servers],
            "externalDocs": asdict(self.info.externalDocs) if self.info.externalDocs.url else None,
        }

    def add_openapi_path_obj(self, route_type: str, endpoint: str, openapi_name: str, openapi_tags: List[str], handler: Callable):
        """
        Adds the given path to openapi spec

        @param route_type: str the http method as string (get, post ...)
        @param endpoint: str the endpoint to be added
        @param openapi_name: str the name of the endpoint
        @param openapi_tags: List[str] for grouping of endpoints
        @param handler: Callable the handler function for the endpoint
        """

        if self.openapi_file_override:
            return

        query_params = None
        request_body = None
        return_annotation = None

        signature = inspect.signature(handler)
        openapi_description = inspect.getdoc(handler) or ""

        if signature:
            parameters = signature.parameters

            if "query_params" in parameters:
                query_params = parameters["query_params"].default

                if query_params is Signature.empty:
                    query_params = None

            if "body" in parameters:
                request_body = parameters["body"].default

                if request_body is Signature.empty:
                    request_body = None

            # priority to typing
            for parameter in parameters:
                param_annotation = parameters[parameter].annotation

                if inspect.isclass(param_annotation):
                    if issubclass(param_annotation, Body):
                        request_body = param_annotation
                    elif issubclass(param_annotation, QueryParams):
                        query_params = param_annotation

            if signature.return_annotation is not Signature.empty:
                return_annotation = signature.return_annotation

        modified_endpoint, path_obj = self.get_path_obj(
            endpoint, openapi_name, openapi_description, openapi_tags, query_params, request_body, return_annotation
        )

        if modified_endpoint not in self.openapi_spec["paths"]:
            self.openapi_spec["paths"][modified_endpoint] = {}
        self.openapi_spec["paths"][modified_endpoint][route_type] = path_obj

    def add_subrouter_paths(self, subrouter_openapi: "OpenAPI"):
        """
        Adds the subrouter paths to main router's openapi specs

        @param subrouter_openapi: OpenAPI the OpenAPI object of the current subrouter
        """

        if self.openapi_file_override:
            return

        paths = subrouter_openapi.openapi_spec["paths"]

        for path in paths:
            self.openapi_spec["paths"][path] = paths[path]

    def get_path_obj(
        self,
        endpoint: str,
        name: str,
        description: str,
        tags: List[str],
        query_params: Optional[str_typed_dict],
        request_body: Optional[str_typed_dict],
        return_annotation: Optional[str_typed_dict],
    ) -> Tuple[str, dict]:
        """
        Get the "path" openapi object according to spec

        @param endpoint: str the endpoint to be added
        @param name: str the name of the endpoint
        @param description: Optional[str] short description of the endpoint (to be fetched from the endpoint definition by default)
        @param tags: List[str] for grouping of endpoints
        @param query_params: Optional[TypedDict] query params for the function
        @param request_body: Optional[TypedDict] request body for the function
        @param return_annotation: Optional[TypedDict] return type of the endpoint handler

        @return: (str, dict) a tuple containing the endpoint with path params wrapped in braces and the "path" openapi object
        according to spec
        """

        if not description:
            description = "No description provided"

        openapi_path_object: dict = {
            "summary": name,
            "description": description,
            "parameters": [],
            "tags": tags,
        }

        # robyn has paths like /:url/:etc whereas openapi requires path like /{url}/{path}
        # this function is used for converting path params to the required form
        # initialized with endpoint for handling endpoints without path params
        endpoint_with_path_params_wrapped_in_braces = endpoint

        endpoint_path_params_split = endpoint.split(":")

        if len(endpoint_path_params_split) > 1:
            endpoint_without_path_params = endpoint_path_params_split[0]

            endpoint_with_path_params_wrapped_in_braces = (
                endpoint_without_path_params[:-1] if endpoint_without_path_params.endswith("/") else endpoint_without_path_params
            )

            for path_param in endpoint_path_params_split[1:]:
                path_param_name = path_param[:-1] if path_param.endswith("/") else path_param

                openapi_path_object["parameters"].append(
                    {
                        "name": path_param_name,
                        "in": "path",
                        "required": True,
                        "schema": {"type": "string"},
                    }
                )
                endpoint_with_path_params_wrapped_in_braces += "/{" + path_param_name + "}"

        if query_params:
            query_param_annotations = query_params.__annotations__ if query_params is str_typed_dict else typing.get_type_hints(query_params)

            for query_param in query_param_annotations:
                query_param_type = self.get_openapi_type(query_param_annotations[query_param])

                openapi_path_object["parameters"].append(
                    {
                        "name": query_param,
                        "in": "query",
                        "required": False,
                        "schema": {"type": query_param_type},
                    }
                )

        if request_body:
            properties = {}

            request_body_annotations = request_body.__annotations__ if request_body is TypedDict else typing.get_type_hints(request_body)

            for body_item in request_body_annotations:
                properties[body_item] = self.get_schema_object(body_item, request_body_annotations[body_item])

            request_body_object = {
                "content": {
                    "application/json": {
                        "schema": {
                            "type": "object",
                            "properties": properties,
                        }
                    }
                }
            }

            openapi_path_object["requestBody"] = request_body_object

        response_schema: dict = {}
        response_type = "text/plain"

        if return_annotation and return_annotation is not Response:
            response_type = "application/json"
            response_schema = self.get_schema_object("response object", return_annotation)

        openapi_path_object["responses"] = {"200": {"description": "Successful Response", "content": {response_type: {"schema": response_schema}}}}

        return endpoint_with_path_params_wrapped_in_braces, openapi_path_object

    def get_openapi_type(self, typed_dict: str_typed_dict) -> str:
        """
        Get actual type from the TypedDict annotations

        @param typed_dict: TypedDict The TypedDict to be converted
        @return: str the type inferred
        """
        type_mapping = {
            int: "integer",
            str: "string",
            bool: "boolean",
            float: "number",
            dict: "object",
            list: "array",
        }

        for type_name in type_mapping:
            if typed_dict is type_name:
                return type_mapping[type_name]

        # default to "string" if type is not found
        return "string"

    def get_schema_object(self, parameter: str, param_type: Any) -> dict:
        """
        Get the schema object for request/response body

        @param parameter: name of the parameter
        @param param_type: Any the type to be inferred
        @return: dict the properties object
        """

        properties: dict = {
            "title": parameter.capitalize(),
        }

        type_mapping: dict = {
            int: "integer",
            str: "string",
            bool: "boolean",
            float: "number",
            dict: "object",
            list: "array",
        }

        for type_name in type_mapping:
            if param_type is type_name:
                properties["type"] = type_mapping[type_name]
                return properties

        # Check if it's a generic type (like List[Object])
        if hasattr(param_type, "__origin__"):
            if param_type.__origin__ is list or param_type.__origin__ is List:
                properties["type"] = "array"
                # Handle the element type in the list
                if hasattr(param_type, "__args__") and param_type.__args__:
                    item_type = param_type.__args__[0]
                    properties["items"] = self.get_schema_object(f"{parameter}_item", item_type)
                return properties

        # check for Optional type
        if param_type.__module__ == "typing":
            properties["anyOf"] = [{"type": self.get_openapi_type(param_type.__args__[0])}, {"type": "null"}]
            return properties
        # check for custom classes and TypedDicts
        elif inspect.isclass(param_type):
            properties["type"] = "object"

            properties["properties"] = {}

            for e in param_type.__annotations__:
                properties["properties"][e] = self.get_schema_object(e, param_type.__annotations__[e])

        properties["type"] = "object"

        return properties

    def override_openapi(self, openapi_json_spec_path: Path):
        """
        Set a pre-configured OpenAPI spec
        @param openapi_json_spec_path: str the path to the json file
        """
        with open(openapi_json_spec_path) as json_file:
            json_file_content = json.load(json_file)
            self.openapi_spec = dict(json_file_content)
            self.openapi_file_override = True

    def get_openapi_docs_page(self) -> Response:
        """
        Handler to the swagger html page to be deployed to the endpoint `/docs`
        @return: FileResponse the swagger html page
        """
        with resources.open_text("robyn", "swagger.html") as path:
            html_file = path.read()
        return html(html_file)

    def get_openapi_config(self) -> dict:
        """
        Returns the openapi spec as a dict
        @return: dict the openapi spec
        """
        return self.openapi_spec



import logging
from enum import Enum
from typing import Optional


class Colors(Enum):
    BLUE = "\033[94m"
    CYAN = "\033[96m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"


class Logger:
    HEADER = "\033[95m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def _format_msg(
        self,
        msg: str,
        color: Optional[Colors],
        bold: bool,
        underline: bool,
    ):
        result = msg
        if color is not None:
            result = f"{color.value}{result}{Logger.ENDC}"
        if bold:
            result = f"{Logger.BOLD}{result}"
        if underline:
            result = f"{Logger.UNDERLINE}{result}"
        return result

    def error(
        self,
        msg: str,
        *args,
        color: Optional[Colors] = Colors.RED,
        bold: bool = False,
        underline: bool = False,
    ):
        self.logger.error(self._format_msg(msg, color, bold, underline), *args)

    def warn(
        self,
        msg: str,
        *args,
        color: Optional[Colors] = Colors.YELLOW,
        bold: bool = False,
        underline: bool = False,
    ):
        self.logger.warn(self._format_msg(msg, color, bold, underline), *args)

    def info(
        self,
        msg: str,
        *args,
        color: Optional[Colors] = Colors.GREEN,
        bold: bool = False,
        underline: bool = False,
    ):
        self.logger.info(self._format_msg(msg, color, bold, underline), *args)

    def debug(
        self,
        msg: str,
        *args,
        color: Colors = Colors.BLUE,
        bold: bool = False,
        underline: bool = False,
    ):
        self.logger.debug(self._format_msg(msg, color, bold, underline), *args)


logger = Logger()



import pytest

from integration_tests.helpers.http_methods_helpers import get, post


@pytest.mark.benchmark
@pytest.mark.parametrize("function_type", ["sync", "async"])
@pytest.mark.parametrize("type_route", ["split_request_untyped", "split_request_typed"])
def test_split_request_params_get_query_params(session, type_route, function_type):
    r = get(f"/{function_type}/{type_route}/query_params?hello=robyn")
    assert r.json() == {"hello": ["robyn"]}
    r = get(f"/{function_type}/{type_route}/query_params?hello=robyn&a=1&b=2")
    assert r.json() == {"hello": ["robyn"], "a": ["1"], "b": ["2"]}
    r = get(f"/{function_type}/{type_route}/query_params")
    assert r.json() == {}


@pytest.mark.benchmark
@pytest.mark.parametrize("function_type", ["sync", "async"])
@pytest.mark.parametrize("type_route", ["split_request_untyped", "split_request_typed"])
def test_split_request_params_get_headers(session, type_route, function_type):
    r = get(f"/{function_type}/{type_route}/headers")
    assert r.text == "robyn"


@pytest.mark.benchmark
@pytest.mark.parametrize("function_type", ["sync", "async"])
@pytest.mark.parametrize("type_route", ["split_request_untyped", "split_request_typed"])
def test_split_request_params_get_path_params(session, type_route, function_type):
    r = get(f"/{function_type}/{type_route}/path_params/123")
    assert r.json() == {"id": "123"}


@pytest.mark.benchmark
@pytest.mark.parametrize("function_type", ["sync", "async"])
@pytest.mark.parametrize("type_route", ["split_request_untyped", "split_request_typed"])
def test_split_request_params_get_method(session, type_route, function_type):
    r = get(f"/{function_type}/{type_route}/method")
    assert r.text == "GET"


@pytest.mark.benchmark
@pytest.mark.parametrize("function_type", ["sync", "async"])
@pytest.mark.parametrize("type_route", ["split_request_untyped", "split_request_typed"])
def test_split_request_params_get_body(session, type_route, function_type):
    res = post(f"/{function_type}/{type_route}/body", data={"hello": "world"})
    assert res.text == "hello=world"


@pytest.mark.benchmark
@pytest.mark.parametrize("function_type", ["sync", "async"])
@pytest.mark.parametrize("type_route", ["split_request_untyped", "split_request_typed"])
def test_split_request_params_get_combined(session, type_route, function_type):
    res = post(
        f"/{function_type}/{type_route}/combined?hello=robyn&a=1&b=2",
        data={"hello": "world"},
    )
    out = res.json()
    assert out["query_params"] == {"hello": ["robyn"], "a": ["1"], "b": ["2"]}
    assert out["body"] == "hello=world"
    assert out["method"] == "POST"
    assert out["url"] == f"/{function_type}/{type_route}/combined"
    assert out["headers"] == "robyn"


@pytest.mark.benchmark
@pytest.mark.parametrize("function_type", ["sync", "async"])
def test_split_request_params_typed_untyped_post_combined(session, function_type):
    res = post(
        f"/{function_type}/split_request_typed_untyped/combined?hello=robyn&a=1&b=2",
        data={"hello": "world"},
    )
    out = res.json()
    assert out["query_params"] == {"hello": ["robyn"], "a": ["1"], "b": ["2"]}
    assert out["body"] == "hello=world"
    assert out["method"] == "POST"
    assert out["url"] == f"/{function_type}/split_request_typed_untyped/combined"
    assert out["headers"] == "robyn"


@pytest.mark.benchmark
@pytest.mark.parametrize("function_type", ["sync", "async"])
def test_split_request_params_get_combined_failure(session, function_type):
    res = post(f"/{function_type}/split_request_typed_untyped/combined/failure?hello=robyn&a=1&b=2", data={"hello": "world"}, should_check_response=False)
    assert 500 == res.status_code



from sqlalchemy import Boolean, Column, Integer, String, create_engine
from sqlalchemy.orm import declarative_base, sessionmaker

Base = declarative_base()

engine = create_engine("sqlite+pysqlite:///:memory:", echo=True)
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)


class User(Base):
    __tablename__ = "users"

    id = Column(Integer, primary_key=True, index=True)
    username = Column(String, unique=True, index=True)
    hashed_password = Column(String)
    is_active = Column(Boolean, default=True)
    is_superuser = Column(Boolean, default=False)


if __name__ == "__main__":
    Base.metadata.create_all(bind=engine)



from unittest.mock import MagicMock, patch

from robyn.cli import create_robyn_app, docs, run, start_app_normally, start_dev_server


# Unit tests
def test_create_robyn_app():
    with patch("robyn.cli.prompt") as mock_prompt:
        mock_prompt.return_value = {
            "directory": "test_dir",
            "docker": "N",
            "project_type": "no-db",
        }
        with patch("robyn.cli.os.makedirs") as mock_makedirs:
            with patch("robyn.cli.shutil.copytree") as mock_copytree, patch("robyn.os.remove") as _mock_remove:
                create_robyn_app()
                mock_makedirs.assert_called_once()
                mock_copytree.assert_called_once()


def test_docs():
    with patch("robyn.cli.webbrowser.open") as mock_open:
        docs()
        mock_open.assert_called_once_with("https://robyn.tech")


def test_start_dev_server():
    config = MagicMock()
    config.dev = True
    with patch("robyn.cli.setup_reloader") as mock_setup_reloader:
        start_dev_server(config, "test_file.py")
        mock_setup_reloader.assert_called_once()


def test_start_app_normally():
    config = MagicMock()
    config.dev = False
    config.parser.parse_known_args.return_value = (MagicMock(), [])
    with patch("robyn.cli.subprocess.run") as mock_run:
        start_app_normally(config)
        mock_run.assert_called_once()


# Integration tests
def test_run_create():
    with patch("robyn.cli.Config") as mock_config:
        mock_config.return_value.create = True
        with patch("robyn.cli.create_robyn_app") as mock_create:
            run()
            mock_create.assert_called_once()



import inspect
import logging
from abc import ABC, abstractmethod
from asyncio import iscoroutinefunction
from functools import wraps
from types import CoroutineType
from typing import Callable, Dict, List, NamedTuple, Optional, Union

from robyn import status_codes
from robyn.authentication import AuthenticationHandler, AuthenticationNotConfiguredError
from robyn.dependency_injection import DependencyMap
from robyn.jsonify import jsonify
from robyn.openapi import OpenAPI
from robyn.responses import FileResponse
from robyn.robyn import FunctionInfo, Headers, HttpMethod, Identity, MiddlewareType, QueryParams, Request, Response, Url
from robyn.types import Body, Files, FormData, IPAddress, Method, PathParams
from robyn.ws import WebSocket

_logger = logging.getLogger(__name__)


def lower_http_method(method: HttpMethod):
    return (str(method))[11:].lower()


class Route(NamedTuple):
    route_type: HttpMethod
    route: str
    function: FunctionInfo
    is_const: bool
    auth_required: bool
    openapi_name: str
    openapi_tags: List[str]


class RouteMiddleware(NamedTuple):
    middleware_type: MiddlewareType
    route: str
    function: FunctionInfo


class GlobalMiddleware(NamedTuple):
    middleware_type: MiddlewareType
    function: FunctionInfo


class BaseRouter(ABC):
    @abstractmethod
    def add_route(*args) -> Union[Callable, CoroutineType, WebSocket]: ...


class Router(BaseRouter):
    def __init__(self) -> None:
        super().__init__()
        self.routes: List[Route] = []

    def _format_tuple_response(self, res: tuple) -> Response:
        if len(res) != 3:
            raise ValueError("Tuple should have 3 elements")

        description, headers, status_code = res
        description = self._format_response(description).description
        new_headers: Headers = Headers(headers)
        if new_headers.contains("Content-Type"):
            headers.set("Content-Type", new_headers.get("Content-Type"))

        return Response(
            status_code=status_code,
            headers=headers,
            description=description,
        )

    def _format_response(
        self,
        res: Union[Dict, Response, bytes, tuple, str],
    ) -> Response:
        if isinstance(res, Response):
            return res

        if isinstance(res, dict):
            return Response(
                status_code=status_codes.HTTP_200_OK,
                headers=Headers({"Content-Type": "application/json"}),
                description=jsonify(res),
            )

        if isinstance(res, FileResponse):
            response: Response = Response(
                status_code=res.status_code,
                headers=res.headers,
                description=res.file_path,
            )
            response.file_path = res.file_path
            return response

        if isinstance(res, bytes):
            return Response(
                status_code=status_codes.HTTP_200_OK,
                headers=Headers({"Content-Type": "application/octet-stream"}),
                description=res,
            )

        if isinstance(res, tuple):
            return self._format_tuple_response(tuple(res))

        return Response(
            status_code=status_codes.HTTP_200_OK,
            headers=Headers({"Content-Type": "text/plain"}),
            description=str(res).encode("utf-8"),
        )

    def add_route(  # type: ignore
        self,
        route_type: HttpMethod,
        endpoint: str,
        handler: Callable,
        is_const: bool,
        auth_required: bool,
        openapi_name: str,
        openapi_tags: List[str],
        exception_handler: Optional[Callable],
        injected_dependencies: dict,
    ) -> Union[Callable, CoroutineType]:
        def wrapped_handler(*args, **kwargs):
            # In the execute functions the request is passed into *args
            request = next(filter(lambda it: isinstance(it, Request), args), None)

            handler_params = inspect.signature(handler).parameters

            if not request or (len(handler_params) == 1 and next(iter(handler_params)) is Request):
                return handler(*args, **kwargs)

            type_mapping = {
                "request": Request,
                "query_params": QueryParams,
                "headers": Headers,
                "path_params": PathParams,
                "body": Body,
                "method": Method,
                "url": Url,
                "form_data": FormData,
                "files": Files,
                "ip_addr": IPAddress,
                "identity": Identity,
            }

            type_filtered_params = {}

            for handler_param in iter(handler_params):
                for type_name in type_mapping:
                    handler_param_type = handler_params[handler_param].annotation
                    handler_param_name = handler_params[handler_param].name
                    if handler_param_type is Request:
                        type_filtered_params[handler_param_name] = request
                    elif handler_param_type is type_mapping[type_name]:
                        type_filtered_params[handler_param_name] = getattr(request, type_name)
                    elif inspect.isclass(handler_param_type):
                        if issubclass(handler_param_type, Body):
                            type_filtered_params[handler_param_name] = getattr(request, "body")
                        elif issubclass(handler_param_type, QueryParams):
                            type_filtered_params[handler_param_name] = getattr(request, "query_params")

            request_components = {
                "r": request,
                "req": request,
                "request": request,
                "query_params": request.query_params,
                "headers": request.headers,
                "path_params": request.path_params,
                "body": request.body,
                "method": request.method,
                "url": request.url,
                "ip_addr": request.ip_addr,
                "identity": request.identity,
                "form_data": request.form_data,
                "files": request.files,
                "router_dependencies": injected_dependencies["router_dependencies"],
                "global_dependencies": injected_dependencies["global_dependencies"],
                **kwargs,
            }

            name_filtered_params = {k: v for k, v in request_components.items() if k in handler_params and k not in type_filtered_params}

            filtered_params = dict(**type_filtered_params, **name_filtered_params)

            if len(filtered_params) != len(handler_params):
                invalid_args = set(handler_params) - set(filtered_params)
                raise SyntaxError(f"Unexpected request params found: {invalid_args}")

            return handler(**filtered_params)

        @wraps(handler)
        async def async_inner_handler(*args, **kwargs):
            try:
                response = self._format_response(
                    await wrapped_handler(*args, **kwargs),
                )
            except Exception as err:
                if exception_handler is None:
                    raise
                response = self._format_response(
                    exception_handler(err),
                )
            return response

        @wraps(handler)
        def inner_handler(*args, **kwargs):
            try:
                response = self._format_response(
                    wrapped_handler(*args, **kwargs),
                )
            except Exception as err:
                if exception_handler is None:
                    raise
                response = self._format_response(
                    exception_handler(err),
                )
            return response

        # these are the arguments
        params = dict(inspect.signature(handler).parameters)

        new_injected_dependencies = {}
        for dependency in injected_dependencies:
            if dependency in params:
                new_injected_dependencies[dependency] = injected_dependencies[dependency]
            else:
                _logger.debug(f"Dependency {dependency} is not used in the handler {handler.__name__}")

        if iscoroutinefunction(handler):
            function = FunctionInfo(
                async_inner_handler,
                True,
                len(params),
                params,
                new_injected_dependencies,
            )
            self.routes.append(Route(route_type, endpoint, function, is_const, auth_required, openapi_name, openapi_tags))
            return async_inner_handler
        else:
            function = FunctionInfo(
                inner_handler,
                False,
                len(params),
                params,
                new_injected_dependencies,
            )
            self.routes.append(Route(route_type, endpoint, function, is_const, auth_required, openapi_name, openapi_tags))
            return inner_handler

    def prepare_routes_openapi(self, openapi: OpenAPI, included_routers: List) -> None:
        for route in self.routes:
            openapi.add_openapi_path_obj(lower_http_method(route.route_type), route.route, route.openapi_name, route.openapi_tags, route.function.handler)

        # TODO! after include_routes does not immediately merge all the routes
        # for router in included_routers:
        #    for route in router:
        #        openapi.add_openapi_path_obj(lower_http_method(route.route_type), route.route, route.openapi_name, route.openapi_tags, route.function.handler)

    def get_routes(self) -> List[Route]:
        return self.routes


class MiddlewareRouter(BaseRouter):
    def __init__(self, dependencies: DependencyMap = DependencyMap()) -> None:
        super().__init__()
        self.global_middlewares: List[GlobalMiddleware] = []
        self.route_middlewares: List[RouteMiddleware] = []
        self.authentication_handler: Optional[AuthenticationHandler] = None
        self.dependencies = dependencies

    def set_authentication_handler(self, authentication_handler: AuthenticationHandler):
        self.authentication_handler = authentication_handler

    def add_route(  # type: ignore
        self,
        middleware_type: MiddlewareType,
        endpoint: str,
        handler: Callable,
        injected_dependencies: dict,
    ) -> Callable:
        params = dict(inspect.signature(handler).parameters)

        new_injected_dependencies = {}
        for dependency in injected_dependencies:
            if dependency in params:
                new_injected_dependencies[dependency] = injected_dependencies[dependency]
            else:
                _logger.debug(f"Dependency {dependency} is not used in the middleware handler {handler.__name__}")

        function = FunctionInfo(
            handler,
            iscoroutinefunction(handler),
            len(params),
            params,
            new_injected_dependencies,
        )
        self.route_middlewares.append(RouteMiddleware(middleware_type, endpoint, function))
        return handler

    def add_auth_middleware(self, endpoint: str):
        """
        This method adds an authentication middleware to the specified endpoint.
        """

        injected_dependencies: dict = {}

        def decorator(handler):
            @wraps(handler)
            def inner_handler(request: Request, *args):
                if not self.authentication_handler:
                    raise AuthenticationNotConfiguredError()
                identity = self.authentication_handler.authenticate(request)
                if identity is None:
                    return self.authentication_handler.unauthorized_response
                request.identity = identity

                return request

            self.add_route(
                MiddlewareType.BEFORE_REQUEST,
                endpoint,
                inner_handler,
                injected_dependencies,
            )
            return inner_handler

        return decorator

    # These inner functions are basically a wrapper around the closure(decorator) being returned.
    # They take a handler, convert it into a closure and return the arguments.
    # Arguments are returned as they could be modified by the middlewares.
    def add_middleware(self, middleware_type: MiddlewareType, endpoint: Optional[str]) -> Callable[..., None]:
        # no dependency injection here
        injected_dependencies: dict = {}

        def inner(handler):
            @wraps(handler)
            async def async_inner_handler(*args, **kwargs):
                return await handler(*args, **kwargs)

            @wraps(handler)
            def inner_handler(*args, **kwargs):
                return handler(*args, **kwargs)

            if endpoint is not None:
                if iscoroutinefunction(handler):
                    self.add_route(
                        middleware_type,
                        endpoint,
                        async_inner_handler,
                        injected_dependencies,
                    )
                else:
                    self.add_route(middleware_type, endpoint, inner_handler, injected_dependencies)
            else:
                params = dict(inspect.signature(handler).parameters)

                if iscoroutinefunction(handler):
                    self.global_middlewares.append(
                        GlobalMiddleware(
                            middleware_type,
                            FunctionInfo(
                                async_inner_handler,
                                True,
                                len(params),
                                params,
                                injected_dependencies,
                            ),
                        )
                    )
                else:
                    self.global_middlewares.append(
                        GlobalMiddleware(
                            middleware_type,
                            FunctionInfo(
                                inner_handler,
                                False,
                                len(params),
                                params,
                                injected_dependencies,
                            ),
                        )
                    )

        return inner

    def get_route_middlewares(self) -> List[RouteMiddleware]:
        return self.route_middlewares

    def get_global_middlewares(self) -> List[GlobalMiddleware]:
        return self.global_middlewares


class WebSocketRouter(BaseRouter):
    def __init__(self) -> None:
        super().__init__()
        self.routes: dict = {}

    def add_route(self, endpoint: str, web_socket: WebSocket) -> None:  # type: ignore
        self.routes[endpoint] = web_socket

    def get_routes(self) -> Dict[str, WebSocket]:
        return self.routes



from typing import Optional

from sqlmodel import Field, SQLModel


class Hero(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    name: str
    secret_name: str
    age: Optional[int] = None



import glob
import os
import signal
import subprocess
import sys
import time
from typing import List, Union

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from robyn.logger import Colors, logger


def compile_rust_files(directory_path: str) -> List[str]:
    rust_files = glob.glob(os.path.join(directory_path, "**/*.rs"), recursive=True)
    rust_binaries: list[str] = []

    for rust_file in rust_files:
        print(f"Compiling rust file: {rust_file}")

        result = subprocess.run(
            [sys.executable, "-m", "rustimport", "build", rust_file],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            start_new_session=False,
        )
        if result.returncode != 0:
            print(f"Error compiling rust file: {rust_file} \n {result.stderr.decode('utf-8')} \n {result.stdout.decode('utf-8')}")
        else:
            print(f"Compiled rust file: {rust_file}")
            rust_file_base = rust_file.removesuffix(".rs")

            # Define the search pattern for the binary file
            if sys.platform == "win32":
                binary_extension = ".dll"
            elif sys.platform == "darwin":
                binary_extension = ".so"
            elif sys.platform == "linux":
                binary_extension = ".so"
            else:
                raise ValueError(f"Unsupported platform: {sys.platform}")

            search_pattern = f"{rust_file_base}.*{binary_extension}"
            # Use glob to find matching binary files
            matching_binaries = glob.glob(search_pattern)
            rust_binaries.extend(matching_binaries)

    return rust_binaries


def create_rust_file(file_name: str) -> None:
    if file_name.endswith(".rs"):
        file_name = file_name.removesuffix(".rs")

    rust_file = f"{file_name}.rs"

    result = subprocess.run(
        [sys.executable, "-m", "rustimport", "new", rust_file],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        start_new_session=False,
    )

    if result.returncode != 0:
        print(
            "Error creating rust file : %s %s",
            result.stderr.decode("utf-8"),
            result.stdout.decode("utf-8"),
        )
    else:
        print("Created rust file : %s", rust_file)


def clean_rust_binaries(rust_binaries: List[str]) -> None:
    for file in rust_binaries:
        print("Cleaning rust file : %s", file)
        os.remove(file)


def setup_reloader(directory_path: str, file_path: str) -> None:
    event_handler = EventHandler(file_path, directory_path)

    # sets the IS_RELOADER_RUNNING environment variable to True
    event_handler.reload()

    logger.info(
        "Dev server initialized with the directory_path : %s",
        directory_path,
        color=Colors.BLUE,
    )

    def terminating_signal_handler(_sig, _frame):
        event_handler.stop_server()
        logger.info("Terminating reloader", bold=True)
        observer.stop()
        observer.join()

    signal.signal(signal.SIGINT, terminating_signal_handler)
    signal.signal(signal.SIGTERM, terminating_signal_handler)

    observer = Observer()
    observer.schedule(event_handler, path=directory_path, recursive=True)
    observer.start()

    try:
        while observer.is_alive():
            observer.join(1)
    finally:
        observer.stop()
        observer.join()
        event_handler.process.wait()


class EventHandler(FileSystemEventHandler):
    def __init__(self, file_path: str, directory_path: str) -> None:
        self.file_path = file_path
        self.directory_path = directory_path
        self.process: Union[subprocess.Popen[bytes], None] = None  # Keep track of the subprocess
        self.built_rust_binaries: List = []  # Keep track of the built rust binaries

        self.last_reload = time.time()  # Keep track of the last reload. EventHandler is initialized with the process.

    def stop_server(self) -> None:
        if self.process:
            os.kill(self.process.pid, signal.SIGTERM)  # Stop the subprocess using os.kill()

    def reload(self) -> None:
        self.stop_server()
        print("Reloading the server")

        new_env = os.environ.copy()
        new_env["IS_RELOADER_RUNNING"] = "True"  # This is used to check if a reloader is already running
        # IS_RELOADER_RUNNING is specifically used for IPC between the reloader and the server

        arguments = [arg for arg in sys.argv[1:] if not arg.startswith("--dev")]

        clean_rust_binaries(self.built_rust_binaries)
        self.built_rust_binaries = compile_rust_files(self.directory_path)

        prev_process = self.process
        if prev_process:
            prev_process.kill()

        self.process = subprocess.Popen(
            [sys.executable, *arguments],
            env=new_env,
        )

        self.last_reload = time.time()

    def on_modified(self, event) -> None:
        """
        This function is a callback that will start a new server on every even change

        :param event FSEvent: a data structure with info about the events
        """

        # Avoid reloading multiple times when watchdog detects multiple events
        if time.time() - self.last_reload < 0.5:
            return

        time.sleep(0.2)  # Wait for the file to be fully written
        self.reload()



import pytest

from integration_tests.helpers.http_methods_helpers import multipart_post


@pytest.mark.benchmark
@pytest.mark.parametrize("function_type", ["sync"])
def test_form_data(function_type: str, session):
    res = multipart_post(f"/{function_type}/form_data", files={"hello": "world"})
    assert "multipart" in res.text


@pytest.mark.benchmark
@pytest.mark.parametrize("function_type", ["sync"])
def test_multipart_file(function_type: str, session):
    res = multipart_post(f"/{function_type}/multipart-file", files={"hello": "world"})
    assert "hello" in res.text



import asyncio
import logging
import os
import socket
from abc import ABC
from pathlib import Path
from typing import Callable, List, Optional, Union

import multiprocess as mp  # type: ignore

from robyn import status_codes
from robyn.argument_parser import Config
from robyn.authentication import AuthenticationHandler
from robyn.dependency_injection import DependencyMap
from robyn.env_populator import load_vars
from robyn.events import Events
from robyn.jsonify import jsonify
from robyn.logger import Colors, logger
from robyn.openapi import OpenAPI
from robyn.processpool import run_processes
from robyn.reloader import compile_rust_files
from robyn.responses import html, serve_file, serve_html
from robyn.robyn import FunctionInfo, Headers, HttpMethod, Request, Response, WebSocketConnector, get_version
from robyn.router import MiddlewareRouter, MiddlewareType, Router, WebSocketRouter
from robyn.types import Directory
from robyn.ws import WebSocket

__version__ = get_version()

config = Config()

if (compile_path := config.compile_rust_path) is not None:
    compile_rust_files(compile_path)
    print("Compiled rust files")


class BaseRobyn(ABC):
    """This is the python wrapper for the Robyn binaries."""

    def __init__(
        self,
        file_object: str,
        config: Config = Config(),
        openapi_file_path: Optional[str] = None,
        openapi: Optional[OpenAPI] = None,
        dependencies: DependencyMap = DependencyMap(),
    ) -> None:
        directory_path = os.path.dirname(os.path.abspath(file_object))
        self.file_path = file_object
        self.directory_path = directory_path
        self.config = config
        self.dependencies = dependencies
        self.openapi = openapi

        self.init_openapi(openapi_file_path)

        if not bool(os.environ.get("ROBYN_CLI", False)):
            # the env variables are already set when are running through the cli
            load_vars(project_root=directory_path)

        self._handle_dev_mode()

        logging.basicConfig(level=self.config.log_level)

        if self.config.log_level.lower() != "warn":
            logger.info(
                "SERVER IS RUNNING IN VERBOSE/DEBUG MODE. Set --log-level to WARN to run in production mode.",
                color=Colors.BLUE,
            )

        self.router = Router()
        self.middleware_router = MiddlewareRouter()
        self.web_socket_router = WebSocketRouter()
        self.request_headers: Headers = Headers({})
        self.response_headers: Headers = Headers({})
        self.excluded_response_headers_paths: Optional[List[str]] = None
        self.directories: List[Directory] = []
        self.event_handlers: dict = {}
        self.exception_handler: Optional[Callable] = None
        self.authentication_handler: Optional[AuthenticationHandler] = None
        self.included_routers: List[Router] = []

    def init_openapi(self, openapi_file_path: Optional[str]) -> None:
        if self.config.disable_openapi:
            return

        if self.openapi is None:
            self.openapi = OpenAPI()

        if openapi_file_path:
            self.openapi.override_openapi(Path(self.directory_path).joinpath(openapi_file_path))
        elif Path(self.directory_path).joinpath("openapi.json").exists():
            self.openapi.override_openapi(Path(self.directory_path).joinpath("openapi.json"))
        # TODO! what about when the elif fails?

    def _handle_dev_mode(self):
        cli_dev_mode = self.config.dev  # --dev
        env_dev_mode = os.getenv("ROBYN_DEV_MODE", "False").lower() == "true"  # ROBYN_DEV_MODE=True
        is_robyn = os.getenv("ROBYN_CLI", False)

        if cli_dev_mode and not is_robyn:
            raise SystemExit("Dev mode is not supported in the python wrapper. Please use the Robyn CLI. e.g. python3 -m robyn app.py --dev")

        if env_dev_mode and not is_robyn:
            logger.error("Ignoring ROBYN_DEV_MODE environment variable. Dev mode is not supported in the python wrapper.")
            raise SystemExit("Dev mode is not supported in the python wrapper. Please use the Robyn CLI. e.g. python3 -m robyn app.py")

    def add_route(
        self,
        route_type: Union[HttpMethod, str],
        endpoint: str,
        handler: Callable,
        is_const: bool = False,
        auth_required: bool = False,
        openapi_name: str = "",
        openapi_tags: Union[List[str], None] = None,
    ):
        """
        Connect a URI to a handler

        :param route_type str: route type between GET/POST/PUT/DELETE/PATCH/HEAD/OPTIONS/TRACE
        :param endpoint str: endpoint for the route added
        :param handler function: represents the sync or async function passed as a handler for the route
        :param is_const bool: represents if the handler is a const function or not
        :param auth_required bool: represents if the route needs authentication or not
        """

        """ We will add the status code here only
        """
        injected_dependencies = self.dependencies.get_dependency_map(self)

        list_openapi_tags: List[str] = openapi_tags if openapi_tags else []

        if auth_required:
            self.middleware_router.add_auth_middleware(endpoint)(handler)

        if isinstance(route_type, str):
            http_methods = {
                "GET": HttpMethod.GET,
                "POST": HttpMethod.POST,
                "PUT": HttpMethod.PUT,
                "DELETE": HttpMethod.DELETE,
                "PATCH": HttpMethod.PATCH,
                "HEAD": HttpMethod.HEAD,
                "OPTIONS": HttpMethod.OPTIONS,
            }
            route_type = http_methods[route_type]

        add_route_response = self.router.add_route(
            route_type=route_type,
            endpoint=endpoint,
            handler=handler,
            is_const=is_const,
            auth_required=auth_required,
            openapi_name=openapi_name,
            openapi_tags=list_openapi_tags,
            exception_handler=self.exception_handler,
            injected_dependencies=injected_dependencies,
        )

        logger.info("Added route %s %s", route_type, endpoint)

        return add_route_response

    def inject(self, **kwargs):
        """
        Injects the dependencies for the route

        :param kwargs dict: the dependencies to be injected
        """
        self.dependencies.add_router_dependency(self, **kwargs)

    def inject_global(self, **kwargs):
        """
        Injects the dependencies for the global routes
        Ideally, this function should be a global function

        :param kwargs dict: the dependencies to be injected
        """
        self.dependencies.add_global_dependency(**kwargs)

    def before_request(self, endpoint: Optional[str] = None) -> Callable[..., None]:
        """
        You can use the @app.before_request decorator to call a method before routing to the specified endpoint

        :param endpoint str|None: endpoint to server the route. If None, the middleware will be applied to all the routes.
        """

        return self.middleware_router.add_middleware(MiddlewareType.BEFORE_REQUEST, endpoint)

    def after_request(self, endpoint: Optional[str] = None) -> Callable[..., None]:
        """
        You can use the @app.after_request decorator to call a method after routing to the specified endpoint

        :param endpoint str|None: endpoint to server the route. If None, the middleware will be applied to all the routes.
        """
        return self.middleware_router.add_middleware(MiddlewareType.AFTER_REQUEST, endpoint)

    def serve_directory(
        self,
        route: str,
        directory_path: str,
        index_file: Optional[str] = None,
        show_files_listing: bool = False,
    ):
        """
        Serves a directory at the given route

        :param route str: the route at which the directory is to be served
        :param directory_path str: the path of the directory to be served
        :param index_file str|None: the index file to be served
        :param show_files_listing bool: if the files listing should be shown or not
        """
        self.directories.append(Directory(route, directory_path, show_files_listing, index_file))

    def add_request_header(self, key: str, value: str) -> None:
        self.request_headers.append(key, value)

    def add_response_header(self, key: str, value: str) -> None:
        self.response_headers.append(key, value)

    def set_request_header(self, key: str, value: str) -> None:
        self.request_headers.set(key, value)

    def set_response_header(self, key: str, value: str) -> None:
        self.response_headers.set(key, value)

    def exclude_response_headers_for(self, excluded_response_headers_paths: Optional[List[str]]):
        """
        To exclude response headers from certain routes
        @param exclude_paths: the paths to exclude response headers from
        """
        self.excluded_response_headers_paths = excluded_response_headers_paths

    def add_web_socket(self, endpoint: str, ws: WebSocket) -> None:
        self.web_socket_router.add_route(endpoint, ws)

    def _add_event_handler(self, event_type: Events, handler: Callable) -> None:
        logger.info("Added event %s handler", event_type)
        if event_type not in {Events.STARTUP, Events.SHUTDOWN}:
            return

        is_async = asyncio.iscoroutinefunction(handler)
        self.event_handlers[event_type] = FunctionInfo(handler, is_async, 0, {}, {})

    def startup_handler(self, handler: Callable) -> None:
        self._add_event_handler(Events.STARTUP, handler)

    def shutdown_handler(self, handler: Callable) -> None:
        self._add_event_handler(Events.SHUTDOWN, handler)

    def is_port_in_use(self, port: int) -> bool:
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                return s.connect_ex(("localhost", port)) == 0
        except Exception:
            raise Exception(f"Invalid port number: {port}")

    def _add_openapi_routes(self, auth_required: bool = False):
        if self.config.disable_openapi:
            return

        if self.openapi is None:
            logger.error("No openAPI")
            return

        self.router.prepare_routes_openapi(self.openapi, self.included_routers)

        self.add_route(
            route_type=HttpMethod.GET,
            endpoint="/openapi.json",
            handler=self.openapi.get_openapi_config,
            is_const=True,
            auth_required=auth_required,
        )
        self.add_route(
            route_type=HttpMethod.GET,
            endpoint="/docs",
            handler=self.openapi.get_openapi_docs_page,
            is_const=True,
            auth_required=auth_required,
        )
        self.exclude_response_headers_for(["/docs", "/openapi.json"])

    def exception(self, exception_handler: Callable):
        self.exception_handler = exception_handler

    def get(
        self,
        endpoint: str,
        const: bool = False,
        auth_required: bool = False,
        openapi_name: str = "",
        openapi_tags: List[str] = ["get"],
    ):
        """
        The @app.get decorator to add a route with the GET method

        :param endpoint str: endpoint for the route added
        :param const bool: represents if the handler is a const function or not
        :param auth_required bool: represents if the route needs authentication or not
        :param openapi_name: str -- the name of the endpoint in the openapi spec
        :param openapi_tags: List[str] -- for grouping of endpoints in the openapi spec
        """

        def inner(handler):
            return self.add_route(HttpMethod.GET, endpoint, handler, const, auth_required, openapi_name, openapi_tags)

        return inner

    def post(
        self,
        endpoint: str,
        auth_required: bool = False,
        openapi_name: str = "",
        openapi_tags: List[str] = ["post"],
    ):
        """
        The @app.post decorator to add a route with POST method

        :param endpoint str: endpoint for the route added
        :param auth_required bool: represents if the route needs authentication or not
        :param openapi_name: str -- the name of the endpoint in the openapi spec
        :param openapi_tags: List[str] -- for grouping of endpoints in the openapi spec
        """

        def inner(handler):
            return self.add_route(HttpMethod.POST, endpoint, handler, auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

        return inner

    def put(
        self,
        endpoint: str,
        auth_required: bool = False,
        openapi_name: str = "",
        openapi_tags: List[str] = ["put"],
    ):
        """
        The @app.put decorator to add a get route with PUT method

        :param endpoint str: endpoint for the route added
        :param auth_required bool: represents if the route needs authentication or not
        :param openapi_name: str -- the name of the endpoint in the openapi spec
        :param openapi_tags: List[str] -- for grouping of endpoints in the openapi spec
        """

        def inner(handler):
            return self.add_route(HttpMethod.PUT, endpoint, handler, auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

        return inner

    def delete(
        self,
        endpoint: str,
        auth_required: bool = False,
        openapi_name: str = "",
        openapi_tags: List[str] = ["delete"],
    ):
        """
        The @app.delete decorator to add a route with DELETE method

        :param endpoint str: endpoint for the route added
        :param auth_required bool: represents if the route needs authentication or not
        :param openapi_name: str -- the name of the endpoint in the openapi spec
        :param openapi_tags: List[str] -- for grouping of endpoints in the openapi spec
        """

        def inner(handler):
            return self.add_route(HttpMethod.DELETE, endpoint, handler, auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

        return inner

    def patch(
        self,
        endpoint: str,
        auth_required: bool = False,
        openapi_name: str = "",
        openapi_tags: List[str] = ["patch"],
    ):
        """
        The @app.patch decorator to add a route with PATCH method

        :param endpoint str: endpoint for the route added
        :param auth_required bool: represents if the route needs authentication or not
        :param openapi_name: str -- the name of the endpoint in the openapi spec
        :param openapi_tags: List[str] -- for grouping of endpoints in the openapi spec
        """

        def inner(handler):
            return self.add_route(HttpMethod.PATCH, endpoint, handler, auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

        return inner

    def head(
        self,
        endpoint: str,
        auth_required: bool = False,
        openapi_name: str = "",
        openapi_tags: List[str] = ["head"],
    ):
        """
        The @app.head decorator to add a route with HEAD method

        :param endpoint str: endpoint for the route added
        :param auth_required bool: represents if the route needs authentication or not
        :param openapi_name: str -- the name of the endpoint in the openapi spec
        :param openapi_tags: List[str] -- for grouping of endpoints in the openapi spec
        """

        def inner(handler):
            return self.add_route(HttpMethod.HEAD, endpoint, handler, auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

        return inner

    def options(
        self,
        endpoint: str,
        auth_required: bool = False,
        openapi_name: str = "",
        openapi_tags: List[str] = ["options"],
    ):
        """
        The @app.options decorator to add a route with OPTIONS method

        :param endpoint str: endpoint for the route added
        :param auth_required bool: represents if the route needs authentication or not
        :param openapi_name: str -- the name of the endpoint in the openapi spec
        :param openapi_tags: List[str] -- for grouping of endpoints in the openapi spec
        """

        def inner(handler):
            return self.add_route(HttpMethod.OPTIONS, endpoint, handler, auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

        return inner

    def connect(
        self,
        endpoint: str,
        auth_required: bool = False,
        openapi_name: str = "",
        openapi_tags: List[str] = ["connect"],
    ):
        """
        The @app.connect decorator to add a route with CONNECT method

        :param endpoint str: endpoint for the route added
        :param auth_required bool: represents if the route needs authentication or not
        :param openapi_name: str -- the name of the endpoint in the openapi spec
        :param openapi_tags: List[str] -- for grouping of endpoints in the openapi spec
        """

        def inner(handler):
            return self.add_route(HttpMethod.CONNECT, endpoint, handler, auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

        return inner

    def trace(
        self,
        endpoint: str,
        auth_required: bool = False,
        openapi_name: str = "",
        openapi_tags: List[str] = ["trace"],
    ):
        """
        The @app.trace decorator to add a route with TRACE method

        :param endpoint str: endpoint for the route added
        :param auth_required bool: represents if the route needs authentication or not
        :param openapi_name: str -- the name of the endpoint in the openapi spec
        :param openapi_tags: List[str] -- for grouping of endpoints in the openapi spec
        """

        def inner(handler):
            return self.add_route(HttpMethod.TRACE, endpoint, handler, auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

        return inner

    def include_router(self, router):
        """
        The method to include the routes from another router

        :param router Robyn: the router object to include the routes from
        """
        self.included_routers.append(router)

        self.router.routes.extend(router.router.routes)
        self.middleware_router.global_middlewares.extend(router.middleware_router.global_middlewares)
        self.middleware_router.route_middlewares.extend(router.middleware_router.route_middlewares)

        if not self.config.disable_openapi and self.openapi is not None:
            self.openapi.add_subrouter_paths(self.openapi)

        # extend the websocket routes
        prefix = router.prefix
        for route in router.web_socket_router.routes:
            new_endpoint = f"{prefix}{route}"
            self.web_socket_router.routes[new_endpoint] = router.web_socket_router.routes[route]

        self.dependencies.merge_dependencies(router)

    def configure_authentication(self, authentication_handler: AuthenticationHandler):
        """
        Configures the authentication handler for the application.

        :param authentication_handler: the instance of a class inheriting the AuthenticationHandler base class
        """
        self.authentication_handler = authentication_handler
        self.middleware_router.set_authentication_handler(authentication_handler)


class Robyn(BaseRobyn):
    def start(self, host: str = "127.0.0.1", port: int = 8080, _check_port: bool = True):
        """
        Starts the server

        :param host str: represents the host at which the server is listening
        :param port int: represents the port number at which the server is listening
        :param _check_port bool: represents if the port should be checked if it is already in use
        """

        host = os.getenv("ROBYN_HOST", host)
        port = int(os.getenv("ROBYN_PORT", port))
        open_browser = bool(os.getenv("ROBYN_BROWSER_OPEN", self.config.open_browser))

        if _check_port:
            while self.is_port_in_use(port):
                logger.error("Port %s is already in use. Please use a different port.", port)
                try:
                    port = int(input("Enter a different port: "))
                except Exception:
                    logger.error("Invalid port number. Please enter a valid port number.")
                    continue

        if not self.config.disable_openapi:
            self._add_openapi_routes()
            logger.info("Docs hosted at http://%s:%s/docs", host, port)

        logger.info("Robyn version: %s", __version__)
        logger.info("Starting server at http://%s:%s", host, port)

        mp.allow_connection_pickling()

        run_processes(
            host,
            port,
            self.directories,
            self.request_headers,
            self.router.get_routes(),
            self.middleware_router.get_global_middlewares(),
            self.middleware_router.get_route_middlewares(),
            self.web_socket_router.get_routes(),
            self.event_handlers,
            self.config.workers,
            self.config.processes,
            self.response_headers,
            self.excluded_response_headers_paths,
            open_browser,
        )


class SubRouter(BaseRobyn):
    def __init__(self, file_object: str, prefix: str = "", config: Config = Config(), openapi: OpenAPI = OpenAPI()) -> None:
        super().__init__(file_object=file_object, config=config, openapi=openapi)
        self.prefix = prefix

    def __add_prefix(self, endpoint: str):
        return f"{self.prefix}{endpoint}"

    def get(self, endpoint: str, const: bool = False, auth_required: bool = False, openapi_name: str = "", openapi_tags: List[str] = ["get"]):
        return super().get(endpoint=self.__add_prefix(endpoint), const=const, auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

    def post(self, endpoint: str, auth_required: bool = False, openapi_name: str = "", openapi_tags: List[str] = ["post"]):
        return super().post(endpoint=self.__add_prefix(endpoint), auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

    def put(self, endpoint: str, auth_required: bool = False, openapi_name: str = "", openapi_tags: List[str] = ["put"]):
        return super().put(endpoint=self.__add_prefix(endpoint), auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

    def delete(self, endpoint: str, auth_required: bool = False, openapi_name: str = "", openapi_tags: List[str] = ["delete"]):
        return super().delete(endpoint=self.__add_prefix(endpoint), auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

    def patch(self, endpoint: str, auth_required: bool = False, openapi_name: str = "", openapi_tags: List[str] = ["patch"]):
        return super().patch(endpoint=self.__add_prefix(endpoint), auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

    def head(self, endpoint: str, auth_required: bool = False, openapi_name: str = "", openapi_tags: List[str] = ["head"]):
        return super().head(endpoint=self.__add_prefix(endpoint), auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

    def trace(self, endpoint: str, auth_required: bool = False, openapi_name: str = "", openapi_tags: List[str] = ["trace"]):
        return super().trace(endpoint=self.__add_prefix(endpoint), auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)

    def options(self, endpoint: str, auth_required: bool = False, openapi_name: str = "", openapi_tags: List[str] = ["options"]):
        return super().options(endpoint=self.__add_prefix(endpoint), auth_required=auth_required, openapi_name=openapi_name, openapi_tags=openapi_tags)


def ALLOW_CORS(app: Robyn, origins: Union[List[str], str], headers: Union[List[str], str] = None):
    """
    Configure CORS headers for the application.

    Args:
        app: Robyn application instance
        origins: List of allowed origins or "*" for all origins
        headers: List of allowed headers or "*" for all headers
    """
    # Handle string input for origins
    if isinstance(origins, str):
        origins = [origins]

    default_headers = ["Content-Type", "Authorization"]
    if isinstance(headers, list):
        headers = list(set(default_headers + headers))
        headers = ", ".join(headers)

    @app.before_request()
    def cors_middleware(request):
        origin = request.headers.get("Origin")

        # If specific origins are set, validate the request origin
        if origin and "*" not in origins and origin not in origins:
            return Response(status_code=403, description="", headers={})

        # Handle preflight requests
        if request.method == "OPTIONS":
            return Response(
                status_code=204,
                headers={
                    "Access-Control-Allow-Origin": origin if origin else (origins[0] if origins else "*"),
                    "Access-Control-Allow-Methods": "GET, POST, PUT, DELETE, PATCH, HEAD, OPTIONS",
                    "Access-Control-Allow-Headers": str(headers) if headers else "Content-Type, Authorization",
                    "Access-Control-Allow-Credentials": "true",
                    "Access-Control-Max-Age": "3600",
                },
                description="",
            )

        return request

    # Set default CORS headers for all responses
    if len(origins) == 1:
        app.set_response_header("Access-Control-Allow-Origin", origins[0])
    else:
        # For multiple origins, we'll handle it dynamically in the response
        app.set_response_header("Access-Control-Allow-Origin", "*")

    app.set_response_header("Access-Control-Allow-Methods", "GET, POST, PUT, DELETE, PATCH, HEAD, OPTIONS")
    app.set_response_header("Access-Control-Allow-Headers", str(headers) if headers else "Content-Type, Authorization")
    app.set_response_header("Access-Control-Allow-Credentials", "true")


__all__ = [
    "Robyn",
    "Request",
    "Response",
    "status_codes",
    "jsonify",
    "serve_file",
    "serve_html",
    "html",
    "ALLOW_CORS",
    "SubRouter",
    "AuthenticationHandler",
    "Headers",
    "WebSocketConnector",
    "WebSocket",
]



import pytest

from integration_tests.helpers.http_methods_helpers import get
from robyn import Robyn


@pytest.mark.benchmark
def test_custom_openapi_spec():
    app = Robyn(__file__, openapi_file_path="openapi_config.json")

    openapi_spec = app.openapi.openapi_spec

    assert isinstance(openapi_spec, dict)

    assert "openapi" in openapi_spec
    assert "info" in openapi_spec
    assert "paths" in openapi_spec
    assert "components" in openapi_spec
    assert "servers" in openapi_spec
    assert "externalDocs" in openapi_spec

    assert openapi_spec["info"]["title"] == "Robyn Test API"
    assert openapi_spec["info"]["version"] == "1.0.0"


@pytest.mark.benchmark
def test_docs_handler():
    # should_check_response = False because check_response raises a
    # failure if the global headers are not present in the response
    # provided we are excluding headers for /docs and /openapi.json
    html_response = get("/docs", should_check_response=False)
    assert html_response.status_code == 200


@pytest.mark.benchmark
def test_json_handler():
    openapi_response = get("/openapi.json", should_check_response=False)

    assert openapi_response.status_code == 200

    openapi_spec = openapi_response.json()

    assert isinstance(openapi_spec, dict)
    assert "openapi" in openapi_spec
    assert "info" in openapi_spec
    assert "paths" in openapi_spec
    assert "components" in openapi_spec
    assert "servers" in openapi_spec
    assert "externalDocs" in openapi_spec


@pytest.mark.benchmark
def test_add_openapi_path():
    openapi_response = get("/openapi.json", should_check_response=False)

    assert openapi_response.status_code == 200

    openapi_spec = openapi_response.json()

    assert isinstance(openapi_spec, dict)

    route_type = "get"
    endpoint = "/openapi_test"
    openapi_description = "Get openapi"
    openapi_tags = ["test tag"]

    assert endpoint in openapi_spec["paths"]
    assert route_type in openapi_spec["paths"][endpoint]
    assert openapi_description == openapi_spec["paths"][endpoint][route_type]["description"]
    assert openapi_tags == openapi_spec["paths"][endpoint][route_type]["tags"]


@pytest.mark.benchmark
def test_add_subrouter_paths():
    openapi_response = get("/openapi.json", should_check_response=False)

    assert openapi_response.status_code == 200

    openapi_spec = openapi_response.json()

    assert isinstance(openapi_spec, dict)

    route_type = "post"
    endpoint = "/sub_router/openapi_test"
    openapi_description = "Get subrouter openapi"
    openapi_tags = ["test subrouter tag"]

    assert endpoint in openapi_spec["paths"]
    assert route_type in openapi_spec["paths"][endpoint]
    assert openapi_description == openapi_spec["paths"][endpoint][route_type]["description"]
    assert openapi_tags == openapi_spec["paths"][endpoint][route_type]["tags"]


@pytest.mark.benchmark
def test_openapi_request_body():
    openapi_response = get("/openapi.json", should_check_response=False)

    assert openapi_response.status_code == 200

    openapi_spec = openapi_response.json()

    assert isinstance(openapi_spec, dict)

    route_type = "post"
    endpoint = "/openapi_request_body"

    assert endpoint in openapi_spec["paths"]
    assert route_type in openapi_spec["paths"][endpoint]
    assert "requestBody" in openapi_spec["paths"][endpoint][route_type]
    assert "content" in openapi_spec["paths"][endpoint][route_type]["requestBody"]
    assert "application/json" in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]
    assert "schema" in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]
    assert "properties" in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]

    assert "name" in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]
    assert "description" in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]
    assert "price" in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]
    assert "tax" in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]

    assert "string" == openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]["description"]["type"]
    assert "number" == openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]["price"]["type"]
    assert "number" == openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]["tax"]["type"]

    assert "object" == openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]["name"]["type"]

    assert "first" in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]["name"]["properties"]
    assert "second" in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]["name"]["properties"]
    assert "initial" in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]["name"]["properties"]

    assert (
        "object"
        in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]["name"]["properties"]["initial"][
            "type"
        ]
    )

    assert (
        "is_present"
        in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]["name"]["properties"]["initial"][
            "properties"
        ]
    )
    assert (
        "letter"
        in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]["name"]["properties"]["initial"][
            "properties"
        ]
    )

    assert {"type": "string"} in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]["name"][
        "properties"
    ]["initial"]["properties"]["letter"]["anyOf"]
    assert {"type": "null"} in openapi_spec["paths"][endpoint][route_type]["requestBody"]["content"]["application/json"]["schema"]["properties"]["name"][
        "properties"
    ]["initial"]["properties"]["letter"]["anyOf"]


@pytest.mark.benchmark
def test_openapi_response_body():
    openapi_response = get("/openapi.json", should_check_response=False)

    assert openapi_response.status_code == 200

    openapi_spec = openapi_response.json()

    assert isinstance(openapi_spec, dict)

    route_type = "post"
    endpoint = "/openapi_request_body"

    assert endpoint in openapi_spec["paths"]
    assert route_type in openapi_spec["paths"][endpoint]
    assert "responses" in openapi_spec["paths"][endpoint][route_type]
    assert "200" in openapi_spec["paths"][endpoint][route_type]["responses"]

    assert openapi_spec["paths"][endpoint][route_type]["responses"]["200"]["description"] == "Successful Response"

    assert "content" in openapi_spec["paths"][endpoint][route_type]["responses"]["200"]

    assert "application/json" in openapi_spec["paths"][endpoint][route_type]["responses"]["200"]["content"]
    assert "schema" in openapi_spec["paths"][endpoint][route_type]["responses"]["200"]["content"]["application/json"]
    assert "properties" in openapi_spec["paths"][endpoint][route_type]["responses"]["200"]["content"]["application/json"]["schema"]

    assert "success" in openapi_spec["paths"][endpoint][route_type]["responses"]["200"]["content"]["application/json"]["schema"]["properties"]
    assert "items_changed" in openapi_spec["paths"][endpoint][route_type]["responses"]["200"]["content"]["application/json"]["schema"]["properties"]

    assert (
        "boolean" == openapi_spec["paths"][endpoint][route_type]["responses"]["200"]["content"]["application/json"]["schema"]["properties"]["success"]["type"]
    )

    assert (
        "integer"
        == openapi_spec["paths"][endpoint][route_type]["responses"]["200"]["content"]["application/json"]["schema"]["properties"]["items_changed"]["type"]
    )


@pytest.mark.benchmark
def test_openapi_query_params():
    openapi_response = get("/openapi.json", should_check_response=False)

    assert openapi_response.status_code == 200

    openapi_spec = openapi_response.json()

    assert isinstance(openapi_spec, dict)

    route_type = "post"
    endpoint = "/openapi_request_body"

    assert endpoint in openapi_spec["paths"]
    assert route_type in openapi_spec["paths"][endpoint]
    assert "parameters" in openapi_spec["paths"][endpoint][route_type]

    assert "required" == openapi_spec["paths"][endpoint][route_type]["parameters"][0]["name"]
    assert "query" == openapi_spec["paths"][endpoint][route_type]["parameters"][0]["in"]
    assert {"type": "boolean"} == openapi_spec["paths"][endpoint][route_type]["parameters"][0]["schema"]



import sqlite3

from robyn import Robyn

app = Robyn(__file__)


@app.get("/")
def index():
    # your db name
    conn = sqlite3.connect("example.db")
    cur = conn.cursor()
    cur.execute("DROP TABLE IF EXISTS test")
    cur.execute("CREATE TABLE test(column_1, column_2)")
    res = cur.execute("SELECT name FROM sqlite_master")
    th = res.fetchone()
    table_name = th[0]
    return f"Hello World! {table_name}"


if __name__ == "__main__":
    app.start(host="0.0.0.0", port=8080)
